import gym
import torch
import torch.nn as nn
import numpy as np

# Define the same Policy Network structure
class PolicyNetwork(nn.Module):
    def __init__(self, input_dim, output_dim):
        super(PolicyNetwork, self).__init__()
        self.fc1 = nn.Linear(input_dim, 128)
        self.dropout = nn.Dropout(0.2)
        self.fc2 = nn.Linear(128, output_dim)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x):
        x = torch.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.softmax(self.fc2(x))
        return x

# Load the trained model
def load_model(filepath, env):
    model = PolicyNetwork(env.observation_space.shape[0], env.action_space.n)
    model.load_state_dict(torch.load(filepath))
    model.eval()
    return model

# Evaluate the model for 100 episodes
def evaluate_model(env, model, num_episodes=100, success_threshold=195):
    success_count = 0
    rewards = []

    for episode in range(num_episodes):
        state = env.reset()[0]
        total_reward = 0

        while True:
            state_tensor = torch.tensor(state, dtype=torch.float32)
            action_probs = model(state_tensor)
            action = torch.argmax(action_probs).item()

            next_state, reward, terminated, truncated, _ = env.step(action)
            total_reward += reward
            state = next_state

            if terminated or truncated:
                break

        rewards.append(total_reward)
        if total_reward >= success_threshold:
            success_count += 1

        print(f"Episode {episode+1}: Total Reward = {total_reward}")

    success_rate = success_count / num_episodes * 100
    print(f"\n✅ Success Rate: {success_rate:.2f}% ({success_count}/{num_episodes} episodes)")

    return success_rate, rewards

if __name__ == "__main__":
    env = gym.make("CartPole-v1", render_mode="human")  # Render for visualization
    model = load_model("reinforce_cartpole.pth", env)
    success_rate, _ = evaluate_model(env, model, num_episodes=100)
    env.close()

    # Save success rate for README update
    with open("evaluation_result.txt", "w") as f:
        f.write(f"Success Rate: {success_rate:.2f}%\n")
