import gym
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import matplotlib.pyplot as plt

# Define the Policy Network
class PolicyNetwork(nn.Module):
    def __init__(self, input_dim, output_dim):
        super(PolicyNetwork, self).__init__()
        self.fc1 = nn.Linear(input_dim, 128)
        self.dropout = nn.Dropout(0.2)
        self.fc2 = nn.Linear(128, output_dim)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x):
        x = torch.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.softmax(self.fc2(x))
        return x

# Define the REINFORCE Algorithm
class REINFORCE:
    def __init__(self, env, learning_rate=5e-3, gamma=0.99):
        self.env = env
        self.gamma = gamma
        self.policy = PolicyNetwork(env.observation_space.shape[0], env.action_space.n)
        self.optimizer = optim.Adam(self.policy.parameters(), lr=learning_rate)
        self.episode_rewards = []

    def select_action(self, state):
        state = torch.tensor(state, dtype=torch.float32)
        action_probs = self.policy(state)
        action = torch.multinomial(action_probs, 1).item()
        return action, action_probs[action]

    def train(self, num_episodes=500):
        all_rewards = []  # Store rewards for plotting

        for episode in range(num_episodes):
            state = self.env.reset()[0]
            episode_memory = []
            episode_reward = 0

            while True:
                action, action_prob = self.select_action(state)
                next_state, reward, terminated, truncated, _ = self.env.step(action)
                episode_memory.append((action_prob, reward))
                episode_reward += reward
                state = next_state
                if terminated or truncated:
                    break

            # Compute returns
            returns = []
            discounted_sum = 0
            for _, reward in reversed(episode_memory):
                discounted_sum = reward + self.gamma * discounted_sum
                returns.insert(0, discounted_sum)

            returns = torch.tensor(returns)
            returns = (returns - returns.mean()) / (returns.std() + 1e-8)

            # Compute policy loss
            loss = 0
            for (action_prob, _), G in zip(episode_memory, returns):
                loss -= torch.log(action_prob) * G

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            all_rewards.append(episode_reward)
            print(f"Episode {episode + 1}: Total Reward = {episode_reward}")

        # Save the model
        torch.save(self.policy.state_dict(), "reinforce_cartpole.pth")
        print("Training complete. Model saved!")

        # Plot rewards
        plt.plot(all_rewards)
        plt.xlabel("Episodes")
        plt.ylabel("Total Reward")
        plt.title("Training Progress")
        plt.savefig("reward_plot.png")  # Save the plot
        print("Reward plot saved as reward_plot.png.")

if __name__ == "__main__":
    env = gym.make("CartPole-v1", render_mode=None)  # Change render_mode to "human" for visualization
    agent = REINFORCE(env)
    agent.train(num_episodes=500)
    env.close()
